==> bless-0.3.1.tar.gz: bless-0.3.1/src/ByteBuffer/SegmentCollection.cs <==
/*
 *   Copyright (c) 2004, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
using NUnit.Framework;
using System;

namespace Bless.Buffers {

///<summary>A collection of buffer segments</summary>
public class SegmentCollection {

#region private data
	Util.List list;
	
	// cache
	Util.List.Node cachedNode;
	long cachedMapping;
#endregion
	
#region Properties	
	public Util.List List {
		get { return list; }
	}
	
#endregion
	
	public SegmentCollection() {
		list= new Util.List();
	}
	
	
	private void InvalidateCache() 
	{
		cachedNode=null;
		cachedMapping=0;
	}
	
	private void SetCache(Util.List.Node n, long map) 
	{
		cachedNode=n;
		cachedMapping=map;
	}
	
	/*
	public byte this[int index] {
		
	}*/
	///
	///<summary>Append a segment to the collection</summary>
	///
	public void Append(Segment s) 
	{
		Util.List.Node n=list.Last;
		if (n!=null && n.data!=null) {
			Segment ls=(Segment)n.data;
			if ((s.buffer==ls.buffer) && (s.start==ls.end+1)) {
				ls.end=s.end;
				return;
			}
		}
		list.Append(s);
	}

	///<summary>Inserts a segment after a node 
	///and merges them if possible</summary>
	private Util.List.Node InsertAfter(Util.List.Node n, Segment s) 
	{
		if (n!=null && n.data!=null) {
			Segment ls=(Segment)n.data;
			if ((s.buffer==ls.buffer) && (s.start==ls.end+1)) {
				ls.end=s.end;
				return n;
			}
		}
		return list.InsertAfter(n, s);
	}
	
	///<summary>Inserts a segment before a node 
	///and merges them if possible</summary>
	private Util.List.Node InsertBefore(Util.List.Node n, Segment s) 
	{
		if (n!=null && n.data!=null) {
			Segment ls=(Segment)n.data;
			if ((s.buffer==ls.buffer) && (s.end+1==ls.start)) {
				ls.start=s.start;
				return n;
			}
		}
		return list.InsertBefore(n, s);
	}
	
	///
	///<summary>Find the segment that the given offset is mapped into</summary>
	///
	public Segment FindSegment(long offset, out long OutMapping, out Util.List.Node OutNode) 
	{
		OutMapping=0;
		OutNode=null;
		
		// if this is the first search
		// set up things
		if (cachedNode==null) {
			if (list.First==null)
				return null;
			SetCache(list.First, 0);
		}	
		
		Segment s=(Segment)cachedNode.data;
		long curMapping=cachedMapping;	
		Util.List.Node curNode=cachedNode;
		
		// is the cached node the one we want?
		if (s.Contains(offset, curMapping)==true) {
			OutMapping=curMapping;
			OutNode=curNode;
			return s;
		}
		
		// search towards the beginning
		if (offset < curMapping ) {
			while (curNode.prev!=null) {
				curNode=curNode.prev;
				s=(Segment)curNode.data;
				curMapping -= s.Size ;
				if (s.Contains(offset, curMapping)==true) {
					SetCache(curNode,curMapping); 
					OutMapping=curMapping;
					OutNode=curNode;
					return s;
				}
			}
		}
		else { // search towards the end
			while (curNode.next!=null) {
				curMapping += s.Size;
				curNode=curNode.next;
				s=(Segment)curNode.data;
				if (s.Contains(offset, curMapping)==true) {
					SetCache(curNode,curMapping);
					OutMapping=curMapping;
					OutNode=curNode;
					return s;
				}
			}
		}
		
		// offset not found but return last accessed node info
		OutMapping=curMapping;
		OutNode=curNode;
		return null;
	}
	
	///
	///<summary>Insert (or Append) a segment collection at a given offset</summary>
	/// 
	public void Insert(SegmentCollection sc, long offset) 
	{
		long mapping;
		Util.List.Node node;
		
		Segment s=FindSegment(offset, out mapping, out node);
		
		// offset not found, check if we have to append
		if (s==null) {
			if ( (node==null && offset==0 ) || (node !=null && offset == mapping + (node.data as Segment).Size)) {
				Util.List lst = sc.List;
				int N=lst.Count;
				Util.List.Node n=lst.First;
				for (int i=0;i<N;i++) {
					Append((Segment)n.data);
					n=n.next;
				}
			}
			return;
		}
		
		if (mapping==offset) { // start of segment?
			// insert data from the current node backwards
			Util.List lst = sc.List;
			int N=lst.Count;
			Util.List.Node n=lst.Last;
			for (int i=0;i<N;i++) { 
				node = InsertBefore(node, (Segment)n.data);
				n=n.prev;
			}
			// update cache
			SetCache(node, mapping);
		}
		else { //middle of segment	
			Segment s1=s.SplitAt(offset-mapping);
			list.InsertAfter(node, s1);
			
			Util.List lst = sc.List;
			int N=lst.Count;
			Util.List.Node n=lst.First;
			for (int i=0;i<N;i++) {
				node = InsertAfter(node, (Segment)n.data);
				n=n.next;
			}
			
		}	
	}
	
	///<summary>Delete a range from the collection</summary>
	public SegmentCollection DeleteRange(long pos1, long pos2) 
	{
		long mapping1, mapping2;
		Util.List.Node node1;
		Util.List.Node node2;
		
		// Find the segments of the end points.
		// Search for the ending point first so that we won't
		// have to invalidate the cache later
		Segment s2=FindSegment(pos2, out mapping2, out node2);
		Segment s1=FindSegment(pos1, out mapping1, out node1);
		
		if (s1==null || s2==null)
			return null;
		
		// ending segment == starting segment 
		// needs special handling
#region Special Handling if node1==node2		
		if (ReferenceEquals(node1, node2)) {
			bool remove_flag=false;
			Segment s_f=s1.SplitAt(pos1-mapping1);
			if (s_f==null) {
				s_f=s1;
				remove_flag=true;
			}
			Segment s_l=s_f.SplitAt(pos2-pos1+1);
			if (s_l!=null)
				list.InsertAfter(node1,s_l);
			if (remove_flag)
				list.Remove(node1);
			SegmentCollection s_c= new SegmentCollection();
			s_c.Append(s_f);
			// invalidate cache. 
			// TO-DO: handle this better
			InvalidateCache();			
	
			return s_c;
		}
#endregion		
		// try to split the ending segment	
		Segment sl=s2.SplitAt(pos2-mapping2+1);
		
		// if we can't, this means that pos2 is the 
		// at the end of the ending segment segment
		if (sl==null)
			sl=s2;	// set the whole segment for removal
		else
			list.InsertAfter(node2, sl); 
			
			
		Util.List.Node n=node1.next;
		
		// try to split the starting segment
		Segment sf=s1.SplitAt(pos1-mapping1);
		
		// if we can't, this means that pos1 is 
		// at the beginning of the starting segment 
		if (sf==null) {
			sf=s1;
			list.Remove(node1); // remove the whole segment
			InvalidateCache();
		}
		
		SegmentCollection sc=new SegmentCollection();
		
		// append the first segment 
		sc.Append(sf);
		
		// append to new and remove from old 
		// all segments up to node2
		while(ReferenceEquals(n, node2)==false) {
			sc.Append((Segment)n.data);
			Util.List.Node p=n;
			n=n.next;
			// Remove() must be placed after n.next
			// because it sets n.next=null
			list.Remove(p);	
		}
		// append and remove node2
		list.Remove(n);
		sc.Append((Segment)n.data);
		
		return sc;
		
	}
	
	
}


#region Tests

[TestFixture]
public class SegmentCollectionTests {

	Buffer buf1;
	Buffer buf2;
	
	class MockBuffer: Buffer {
	
		public override void Put(long pos, byte[] data) { }
		public override int Get(byte[] buf, long pos, int len)  { return 0;}
		public override void Append(byte[] data) { }
		public override void Append(byte data) { }
		public override byte this[long index] {
			set { }
			get { return 0;}
		}
		public override long Size {
			get { return 0; }
		}
	
	}
	
	[SetUp]
	public void Init() {
		buf1 = new MockBuffer();
		buf2 = new MockBuffer();
	}
	
	[Test]
	public void AppendSegmentsTest1() {
		SegmentCollection segs = new SegmentCollection();
		Segment s1,s2;
		s1=new Segment(buf1, 0, 5);
		s2=new Segment(buf2, 6, 5);
		segs.Append(s1);
		segs.Append(s2);
		Assert.AreSame(s1, segs.List[0]);
		Assert.AreSame(s2, segs.List[1]);
	}

	[Test]
	public void AppendSegmentsTest2() {
		SegmentCollection segs = new SegmentCollection();
		Segment s1,s2;
		s1=new Segment(buf1, 0, 5);
		s2=new Segment(buf1, 6, 8);
		segs.Append(s1);
		segs.Append(s2);
		Assert.AreSame(s1, segs.List[0]);
		Assert.IsTrue(s1.start==0);
		Assert.IsTrue(s1.end==8);
		//Assert.IsNull(segs[1]);
	}
	
	[Test]
	public void FindSegmentTest() {
		SegmentCollection segs1 = new SegmentCollection();	
		Segment s1,s2,s3,s4;		
		s1=new Segment(buf1, 0, 5);
		s2=new Segment(buf2, 6, 8);
		s3=new Segment(buf2, 0, 5);
		s4=new Segment(buf1, 6, 8);
		segs1.Append(s1);
		segs1.Append(s2);
		segs1.Append(s3);
		segs1.Append(s4);
		
		long map;
		Util.List.Node n;
		
		Assert.AreSame(s1, segs1.FindSegment(4, out map, out n));
		Assert.AreEqual(0, map);
		Assert.AreSame(s3, segs1.FindSegment(10,  out map, out n));
		Assert.AreEqual(9, map);
	}
	
	
	[Test]
	public void InsertSegmentCollectionTest1() {
		SegmentCollection segs1 = new SegmentCollection();
		SegmentCollection segs2 = new SegmentCollection();
		Segment s1,s2,s3,s4;
		s1=new Segment(buf1, 0, 5);
		s2=new Segment(buf2, 6, 8);
		s3=new Segment(buf2, 0, 5);
		s4=new Segment(buf1, 6, 8);
		
		segs1.Append(s1);
		segs1.Append(s2);
		segs2.Append(s3);
		segs2.Append(s4);
		
		segs1.Insert(segs2, 6);
		
		Assert.AreSame(s1, segs1.List[0]);
		Assert.AreSame(s3, segs1.List[1]);
		Assert.AreSame(s4, segs1.List[2]);
		Assert.AreSame(s2, segs1.List[3]);
		//Assert.IsNull(segs[1]);
	}
	
	
	[Test]
	public void InsertSegmentCollectionTest2() {
		SegmentCollection segs1 = new SegmentCollection();
		SegmentCollection segs2 = new SegmentCollection();
		Segment s1,s2,s3,s4;
		s1=new Segment(buf1, 0, 5);
		s2=new Segment(buf2, 6, 8);
		s3=new Segment(buf2, 0, 5);
		s4=new Segment(buf1, 6, 8);
		
		segs1.Append(s1);
		segs1.Append(s2);
		segs2.Append(s3);
		segs2.Append(s4);
		
		segs1.Insert(segs2, 4);
		
		Assert.IsTrue(s1.end==3);
		Assert.AreSame(s1, segs1.List[0]);
		Assert.AreSame(s3, segs1.List[1]);
		Assert.AreSame(s4, segs1.List[2]);
		Assert.AreSame(s2, segs1.List[4]);
		Assert.IsTrue(((Segment)segs1.List[3]).start==4);
		Assert.IsTrue(((Segment)segs1.List[3]).buffer==buf1);
		Assert.IsTrue(((Segment)segs1.List[3]).end==5);
		//Assert.IsNull(segs[1]);
	}
	
	[Test]
	public void InsertSegmentCollectionAtEndTest() {
		SegmentCollection segs1 = new SegmentCollection();
		SegmentCollection segs2 = new SegmentCollection();
		Segment s1,s2,s3,s4;
		s1=new Segment(buf1, 0, 5);
		s2=new Segment(buf2, 6, 8);
		s3=new Segment(buf2, 0, 5);
		s4=new Segment(buf1, 6, 8);
		
		segs1.Append(s1);
		segs1.Append(s2);
		segs2.Append(s3);
		segs2.Append(s4);
		
		segs1.Insert(segs2, 9);
		
		Assert.AreSame(s1, segs1.List[0]);
		Assert.AreSame(s2, segs1.List[1]);
		Assert.AreSame(s3, segs1.List[2]);
		Assert.AreSame(s4, segs1.List[3]);
		//Assert.IsNull(segs[1]);
	}
	
	[Test]
	public void InsertSegmentCollectionThenFindTest() {
		SegmentCollection segs1 = new SegmentCollection();
		SegmentCollection segs2 = new SegmentCollection();
		Segment s1,s2,s3,s4;
		s1=new Segment(buf1, 0, 5);
		s2=new Segment(buf2, 6, 8);
		s3=new Segment(buf2, 0, 5);
		s4=new Segment(buf1, 6, 8);
		
		segs1.Append(s1);
		segs1.Append(s2);
		segs2.Append(s3);
		segs2.Append(s4);
		
		segs1.Insert(segs2, 4);
		
		long map;
		Util.List.Node n;
		//segs1.List.Display();
		Assert.AreSame(s1, segs1.FindSegment(3, out map, out n),"s1");
		Assert.AreEqual(0, map);
		Assert.AreSame(s3, segs1.FindSegment(7,  out map, out n),"s3");
		Assert.AreEqual(4, map);
		Assert.AreSame(s4, segs1.FindSegment(12,  out map, out n),"s4");
		Assert.AreEqual(10, map);
		Segment s5=segs1.FindSegment(13,  out map, out n);
		Assert.AreEqual(13, map);
		Assert.IsTrue(s5.end==s5.start+1,"s1'");
		Assert.AreSame(s2, segs1.FindSegment(15,  out map, out n),"s2");
		Assert.AreEqual(15, map);
		
	}
	
	
	[Test]
	public void DeleteFromSegmentCollectionTest1() {
		SegmentCollection segs1 = new SegmentCollection();
		Segment s1,s2,s3,s4;
		s1=new Segment(buf1, 0, 5);
		s2=new Segment(buf2, 6, 8);
		s3=new Segment(buf2, 0, 5);
		s4=new Segment(buf1, 6, 8);
		segs1.Append(s1);
		segs1.Append(s2);
		segs1.Append(s3);
		segs1.Append(s4);	
		
		SegmentCollection del=segs1.DeleteRange(7,15);
		Assert.AreEqual(3, segs1.List.Count,"#1");
		Assert.AreEqual(6, ((Segment)segs1.List[0]).Size,"#2");
		Assert.AreEqual(6, ((Segment)segs1.List[1]).start,"#3");
		Assert.AreEqual(6, ((Segment)segs1.List[1]).end,"#4");
		Assert.AreEqual(7, ((Segment)segs1.List[2]).start,"#5");
		Assert.AreEqual(8, ((Segment)segs1.List[2]).end,"#6");
	}	
		
	[Test]
	public void DeleteFromSegmentCollectionTest2() {	
		SegmentCollection segs1 = new SegmentCollection();
		Segment s1;
		s1=new Segment(buf1, 0, 8);
		segs1.Append(s1);
		SegmentCollection del=segs1.DeleteRange(3,6);
		
		Assert.AreEqual(2, segs1.List.Count, "#0");
		Assert.AreEqual(0, ((Segment)segs1.List[0]).start,"#1");
		Assert.AreEqual(2, ((Segment)segs1.List[0]).end,"#2");
		Assert.AreEqual(7, ((Segment)segs1.List[1]).start,"#3");
		Assert.AreEqual(8, ((Segment)segs1.List[1]).end,"#4");
		
		Assert.AreEqual(1, del.List.Count, "#5");
		Assert.AreEqual(3, ((Segment)del.List[0]).start,"#6");
		Assert.AreEqual(6, ((Segment)del.List[0]).end,"#7");
		
	}
	
	[Test]
	public void DeleteFromSegmentCollectionTest3() {	
		SegmentCollection segs1 = new SegmentCollection();
		Segment s1;
		s1=new Segment(buf1, 0, 8);
		segs1.Append(s1);
		SegmentCollection del=segs1.DeleteRange(0,6);
		
		Assert.AreEqual(1, segs1.List.Count, "#0");
		Assert.AreEqual(7, ((Segment)segs1.List[0]).start,"#1");
		Assert.AreEqual(8, ((Segment)segs1.List[0]).end,"#2");
				
		Assert.AreEqual(1, del.List.Count, "#5");
		Assert.AreEqual(0, ((Segment)del.List[0]).start,"#6");
		Assert.AreEqual(6, ((Segment)del.List[0]).end,"#7");
		
	}
	
	[Test]
	public void DeleteFromSegmentCollectionTest4() {	
		SegmentCollection segs1 = new SegmentCollection();
		Segment s1;
		s1=new Segment(buf1, 0, 8);
		segs1.Append(s1);
		SegmentCollection del=segs1.DeleteRange(0,8);
		
		Assert.AreEqual(0, segs1.List.Count, "#0");
		
		Assert.AreEqual(1, del.List.Count, "#5");
		Assert.AreEqual(0, ((Segment)del.List[0]).start,"#6");
		Assert.AreEqual(8, ((Segment)del.List[0]).end,"#7");
		
	}
	
	[Test]
	public void DeleteFromSegmentCollectionTest5() {	
		SegmentCollection segs1 = new SegmentCollection();
		Segment s1;
		s1=new Segment(buf1, 0, 8);
		segs1.Append(s1);
		SegmentCollection del=segs1.DeleteRange(6,8);
		
		Assert.AreEqual(1, segs1.List.Count, "#0");
		Assert.AreEqual(0, ((Segment)segs1.List[0]).start,"#1");
		Assert.AreEqual(5, ((Segment)segs1.List[0]).end,"#2");
				
		Assert.AreEqual(1, del.List.Count, "#5");
		Assert.AreEqual(6, ((Segment)del.List[0]).start,"#6");
		Assert.AreEqual(8, ((Segment)del.List[0]).end,"#7");
	}
	
	[Test]
	public void InsertToSegmentCollectionThenDeleteTest() {
		SegmentCollection segs1 = new SegmentCollection();
		Segment s1, s2;
		s1=new Segment(buf1, 0, 2);
		s2=new Segment(buf1, 3, 5);
		segs1.Append(s1);
		SegmentCollection segs2 =new SegmentCollection();
		segs2.Append(s2);
		segs1.Insert(segs2, 2);
		
		segs1.DeleteRange(2,4);
		
		long map;
		Util.List.Node n;
		Segment s=segs1.FindSegment(0,out map, out n);
		Assert.IsNotNull(s,"#1");
		Assert.AreEqual(0, map,"#2");
		Assert.AreEqual(0, s.start,"#3");
		  
		
		
	}
	
	[Test]
	public void ReplaceAtEnd() {
		SegmentCollection segs1 = new SegmentCollection();
		Segment s1, s2, s3;
		s1=new Segment(buf1, 0, 4);
		s2=new Segment(buf2, 7, 7);
		s3=new Segment(buf2, 8, 8);
		segs1.Append(s1);
		//segs1.List.Display();
		segs1.DeleteRange(4,4);
		//segs1.List.Display();
		segs1.Append(s2);
		//segs1.List.Display();
		segs1.DeleteRange(4,4);
		//segs1.List.Display();
		segs1.Append(s3);
		//segs1.List.Display();
		
		Assert.AreEqual(2, segs1.List.Count,"#1");
		Assert.AreEqual(3, (segs1.List[0] as Segment).end,"#2");
		Assert.AreEqual(8, (segs1.List[1] as Segment).start,"#3");
		Assert.AreEqual(8, (segs1.List[1] as Segment).end,"#4");
		
	}
	
	[Test]
	public void InsertInEmptyCollection() {
		SegmentCollection segs1 = new SegmentCollection();
		SegmentCollection segs2 = new SegmentCollection();
		Segment s1=new Segment(buf1, 0, 4);
		Segment s2=new Segment(buf2, 7, 9);
		
		segs2.Append(s1);
		segs2.Append(s2);
		
		segs1.Insert(segs2, 0);
		
		Assert.AreEqual(2, segs1.List.Count,"#1");
		Assert.AreEqual(4, (segs1.List[0] as Segment).end,"#2");
		Assert.AreEqual(7, (segs1.List[1] as Segment).start,"#3");
		Assert.AreEqual(9, (segs1.List[1] as Segment).end,"#4");	
	}	
}
#endregion
} // namespace

==> bless-0.3.4.tar.gz: bless-0.3.4/src/ByteBuffer/SegmentCollection.cs <==
/*
 *   Copyright (c) 2004, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
using System;

namespace Bless.Buffers {

///<summary>A collection of buffer segments</summary>
public class SegmentCollection {

#region private data
	Util.List list;
	
	// cache
	Util.List.Node cachedNode;
	long cachedMapping;
#endregion
	
#region Properties	
	public Util.List List {
		get { return list; }
	}
	
#endregion
	
	public SegmentCollection() {
		list= new Util.List();
	}
	
	
	private void InvalidateCache() 
	{
		cachedNode=null;
		cachedMapping=0;
	}
	
	private void SetCache(Util.List.Node n, long map) 
	{
		cachedNode=n;
		cachedMapping=map;
	}
	
	/*
	public byte this[int index] {
		
	}*/
	///
	///<summary>Append a segment to the collection</summary>
	///
	public void Append(Segment s) 
	{
		Util.List.Node n=list.Last;
		if (n!=null && n.data!=null) {
			Segment ls=(Segment)n.data;
			if ((s.buffer==ls.buffer) && (s.start==ls.end+1)) {
				ls.end=s.end;
				return;
			}
		}
		list.Append(s);
	}

	///<summary>Inserts a segment after a node 
	///and merges them if possible</summary>
	private Util.List.Node InsertAfter(Util.List.Node n, Segment s) 
	{
		if (n!=null && n.data!=null) {
			Segment ls=(Segment)n.data;
			if ((s.buffer==ls.buffer) && (s.start==ls.end+1)) {
				ls.end=s.end;
				return n;
			}
		}
		return list.InsertAfter(n, s);
	}
	
	///<summary>Inserts a segment before a node 
	///and merges them if possible</summary>
	private Util.List.Node InsertBefore(Util.List.Node n, Segment s) 
	{
		if (n!=null && n.data!=null) {
			Segment ls=(Segment)n.data;
			if ((s.buffer==ls.buffer) && (s.end+1==ls.start)) {
				ls.start=s.start;
				return n;
			}
		}
		return list.InsertBefore(n, s);
	}
	
	///
	///<summary>Find the segment that the given offset is mapped into</summary>
	///
	public Segment FindSegment(long offset, out long OutMapping, out Util.List.Node OutNode) 
	{
		OutMapping=0;
		OutNode=null;
		
		// if this is the first search
		// set up things
		if (cachedNode==null) {
			if (list.First==null)
				return null;
			SetCache(list.First, 0);
		}	
		
		Segment s=(Segment)cachedNode.data;
		long curMapping=cachedMapping;	
		Util.List.Node curNode=cachedNode;
		
		// is the cached node the one we want?
		if (s.Contains(offset, curMapping)==true) {
			OutMapping=curMapping;
			OutNode=curNode;
			return s;
		}
		
		// search towards the beginning
		if (offset < curMapping ) {
			while (curNode.prev!=null) {
				curNode=curNode.prev;
				s=(Segment)curNode.data;
				curMapping -= s.Size ;
				if (s.Contains(offset, curMapping)==true) {
					SetCache(curNode,curMapping); 
					OutMapping=curMapping;
					OutNode=curNode;
					return s;
				}
			}
		}
		else { // search towards the end
			while (curNode.next!=null) {
				curMapping += s.Size;
				curNode=curNode.next;
				s=(Segment)curNode.data;
				if (s.Contains(offset, curMapping)==true) {
					SetCache(curNode,curMapping);
					OutMapping=curMapping;
					OutNode=curNode;
					return s;
				}
			}
		}
		
		// offset not found but return last accessed node info
		OutMapping=curMapping;
		OutNode=curNode;
		return null;
	}
	
	///
	///<summary>Insert (or Append) a segment collection at a given offset</summary>
	/// 
	public void Insert(SegmentCollection sc, long offset) 
	{
		long mapping;
		Util.List.Node node;
		
		Segment s=FindSegment(offset, out mapping, out node);
		
		// offset not found, check if we have to append
		if (s==null) {
			if ( (node==null && offset==0 ) || (node !=null && offset == mapping + (node.data as Segment).Size)) {
				Util.List lst = sc.List;
				int N=lst.Count;
				Util.List.Node n=lst.First;
				for (int i=0;i<N;i++) {
					Append((Segment)n.data);
					n=n.next;
				}
			}
			return;
		}
		
		if (mapping==offset) { // start of segment?
			// insert data from the current node backwards
			Util.List lst = sc.List;
			int N=lst.Count;
			Util.List.Node n=lst.Last;
			for (int i=0;i<N;i++) { 
				node = InsertBefore(node, (Segment)n.data);
				n=n.prev;
			}
			// update cache
			SetCache(node, mapping);
		}
		else { //middle of segment	
			Segment s1=s.SplitAt(offset-mapping);
			list.InsertAfter(node, s1);
			
			Util.List lst = sc.List;
			int N=lst.Count;
			Util.List.Node n=lst.First;
			for (int i=0;i<N;i++) {
				node = InsertAfter(node, (Segment)n.data);
				n=n.next;
			}
			
		}	
	}
	
	///<summary>Delete a range from the collection</summary>
	public SegmentCollection DeleteRange(long pos1, long pos2) 
	{
		long mapping1, mapping2;
		Util.List.Node node1;
		Util.List.Node node2;
		
		// Find the segments of the end points.
		// Search for the ending point first so that we won't
		// have to invalidate the cache later
		Segment s2=FindSegment(pos2, out mapping2, out node2);
		Segment s1=FindSegment(pos1, out mapping1, out node1);
		
		if (s1==null || s2==null)
			return null;
		
		// ending segment == starting segment 
		// needs special handling
#region Special Handling if node1==node2		
		if (ReferenceEquals(node1, node2)) {
			bool remove_flag=false;
			
			// try to split segment at pos1
			Segment s_f=s1.SplitAt(pos1-mapping1);
			// if we can't, this means that pos1 is 
			// at the beginning of the segment 
			if (s_f==null) {
				s_f=s1;
				remove_flag=true;
			}
			
			// try to split s_f at pos2+1
			// s_l is the ending part of s1 that we 
			// should keep
			Segment s_l=s_f.SplitAt(pos2-pos1+1);
			
			// if we can't split, this means that pos2 is 
			// at the end of the segment,
			// otherwise add s_l after node1 (which contains s1) 
			if (s_l!=null)
				list.InsertAfter(node1,s_l);
			
			// if we should remove s1
			if (remove_flag) {
				// try to set the cache
				if (node1.next!=null)
					SetCache(node1.next, mapping1);
				else if (node1.prev!=null) {
					Segment s=(Segment)node1.prev.data;
					SetCache(node1.prev, mapping1-s.Size);
				}
				else
					InvalidateCache();
					
				list.Remove(node1);
			}
			//else leave the cache set as is (node1, mapping1)
			
			SegmentCollection s_c= new SegmentCollection();
			s_c.Append(s_f);
			
			return s_c;
		}
#endregion		
		// try to split the ending segment	
		Segment sl=s2.SplitAt(pos2-mapping2+1);
		
		// if we can't, this means that pos2 is the 
		// at the end of the ending segment
		if (sl==null)
			sl=s2;	// set the whole segment for removal
		else
			list.InsertAfter(node2, sl); 
			
			
		Util.List.Node n=node1.next;
		
		// try to split the starting segment
		Segment sf=s1.SplitAt(pos1-mapping1);
		
		// if we can't, this means that pos1 is 
		// at the beginning of the starting segment 
		if (sf==null) {
			sf=s1;
			// try to set the cache
			if (node1.prev!=null) {
				Segment s=(Segment)node1.prev.data;
				SetCache(node1.prev, mapping1-s.Size);
			}
			else 
				InvalidateCache();
			
			list.Remove(node1); // remove the whole segment
		}
		
		SegmentCollection sc=new SegmentCollection();
		
		// append the first segment 
		sc.Append(sf);
		
		// append to new and remove from old 
		// all segments up to node2
		while(ReferenceEquals(n, node2)==false) {
			sc.Append((Segment)n.data);
			Util.List.Node p=n;
			n=n.next;
			// Remove() must be placed after n.next
			// because it sets n.next=null
			list.Remove(p);	
		}
		// append and remove node2
		list.Remove(n);
		sc.Append((Segment)n.data);
		
		return sc;
		
	}
	
	
}

} // end namespace
